import ntpath
import re
from configparser import ConfigParser, ParsingError
from csv import DictReader
from io import StringIO
from pathlib import Path
from typing import Any

from win2xcur.parser import open_blob
from win2xcur.theme import CursorTheme, WIN_CURSORS

reexpand = re.compile(r'%(\w*)%')


def expand_registry(text: str, names: dict[str, str]) -> str:
    def replacer(match: re.Match[str]) -> str:
        name = match.group(1)
        if not name:
            return '%'
        elif name in names:
            return names[name].strip('"')
        else:
            return match.group(0)

    return reexpand.sub(replacer, text)


def parse_inf(inf: Path) -> CursorTheme:
    parser = ConfigParser(allow_no_value=True, strict=False)

    try:
        parser.read(inf)
    except ParsingError as e:
        raise ValueError(e.args[0])

    try:
        reg_section = parser['DefaultInstall']['AddReg']
    except KeyError:
        raise ValueError('Unable to find registry update section in INF')

    try:
        updates = list(parser[reg_section])
    except KeyError:
        raise ValueError(f'Registry update section does not exist in INF: {reg_section}')

    updates = [update for update in updates if update.startswith((
        'hkcu,"control panel\\cursors\\schemes",',
        'hklm,"software\\microsoft\\windows\\currentversion\\control panel\\cursors\\schemes",',
    ))]

    if len(updates) == 0:
        raise ValueError('No cursor installs found in INF')
    elif len(updates) > 1:
        raise ValueError('Multiple cursor installs found in INF')

    try:
        strings = dict(parser['Strings'])
    except KeyError:
        strings = {}

    parsed = next(DictReader(StringIO(updates[0]), fieldnames=['root', 'path', 'name', 'flags', 'value']))

    params: dict[str, Any] = {'name': expand_registry(parsed['name'], strings)}
    cursor_paths = expand_registry(parsed['value'], strings).split(',')

    casefolded = {file.name.casefold(): file for file in inf.parent.iterdir() if file.is_file()}

    for name, filename in zip(WIN_CURSORS, cursor_paths):
        if not filename:
            continue

        basename = ntpath.basename(filename)
        try:
            with open(casefolded[basename.casefold()], 'rb') as f:
                params[name] = open_blob(f.read())
        except FileNotFoundError:
            raise ValueError(f'Expected cursor file in same directory as INF: {basename}')

    return CursorTheme(**params)



from typing import List

from win2xcur.cursor import CursorFrame


def apply_to_frames(frames: List[CursorFrame], *, scale: float) -> None:
    for frame in frames:
        for cursor in frame:
            cursor.image.scale(
                int(round(cursor.image.width * scale)),
                int(round(cursor.image.height * scale)),
            )

            hx, hy = cursor.hotspot
            cursor.hotspot = (int(hx * scale), int(hy * scale))



from dataclasses import dataclass, fields
from typing import Optional

from win2xcur.parser.base import BaseParser


@dataclass
class CursorTheme:
    name: str

    # Standard Windows cursors
    arrow: Optional[BaseParser] = None
    help: Optional[BaseParser] = None
    working: Optional[BaseParser] = None
    wait: Optional[BaseParser] = None
    crosshair: Optional[BaseParser] = None
    text: Optional[BaseParser] = None
    pen: Optional[BaseParser] = None
    unavailable: Optional[BaseParser] = None
    size_ns: Optional[BaseParser] = None
    size_ew: Optional[BaseParser] = None
    size_nwse: Optional[BaseParser] = None
    size_nesw: Optional[BaseParser] = None
    move: Optional[BaseParser] = None
    up_arrow: Optional[BaseParser] = None
    link: Optional[BaseParser] = None
    location: Optional[BaseParser] = None
    person: Optional[BaseParser] = None

    # TODO: add some generated cursors here


ALL_CURSORS = [field.name for field in fields(CursorTheme) if field.name != 'name']
WIN_CURSORS = ALL_CURSORS[:17]

XCURSOR_ALIASES = {
    'arrow': [
        # Actual arrows
        'default', 'arrow', 'left_ptr', 'top_left_arrow',

        # Other directional cursors
        # TODO: consider rotating up-arrow
        'left-arrow', 'right-arrow', 'down-arrow',
        'sb_left_arrow', 'sb_right_arrow', 'sb_down_arrow',

        # Grab
        'grab', 'openhand',

        # Alias/link
        'alias', 'link', 'dnd-link', '3085a0e285430894940527032f8b26df', '640fb0e74195791501fd1ed57b41487f',
        'a2a266d0498c3104214a47bd64ab0fc8',

        # Sides and corners
        'top_left_corner', 'top_right_corner', 'bottom_left_corner', 'bottom_right_corner',
        'top_side', 'bottom_side', 'left_side', 'right_side',
        'ul_angle', 'ur_angle', 'll_angle', 'lr_angle',

        # Right pointer
        # TODO: consider mirroring left cursor
        'right_ptr', 'draft_large', 'draft_small',

        # Vertical text
        # TODO: consider rotating text
        'vertical-text',

        # Copy
        'copy', 'dnd-copy', '1081e37283d90000800003c07f3ef6bf', '6407b0e94181790501fd1e167b474872',
        'b66166c04f8c3109214a4fbd64a50fc8',

        # Zooming
        'zoom-in', 'zoom-out',

        # Dotbox
        'dotbox', 'dot_box_mask', 'draped_box', 'icon', 'target',

        # Miscellaneous cursors
        'context-menu', 'center_ptr', 'color-picker', 'X_cursor', 'x-cursor', 'wayland-cursor', 'pirate',
        'top_tee', 'bottom_tee', 'left_tee', 'right_tee',
    ],
    'help': ['help', 'left_ptr_help', 'question_arrow', 'whats_this', '5c6cd98b3f3ebcb1f9c7f1c204630408',
             'd9ce0ab605698f320427677b458ad60b'],
    'working': ['progress', 'half-busy', 'left_ptr_watch', '00000000000000020006000e7e9ffc3f',
                '08e8e1c95fe2fc01f976f1e063a24ccd', '3ecb610c1bf2410f44200f48c40d3599'],
    'wait': ['wait', 'watch'],
    'crosshair': [
        # Regular crosshair
        'crosshair', 'cross', 'tcross', 'cross_reverse', 'diamond_cross',

        # Cell selector
        'cell', 'plus',
    ],
    'text': ['text', 'ibeam', 'xterm'],
    'pen': ['pencil', 'draft'],
    'unavailable': [
        # Regular unavailable
        'not-allowed', 'circle', 'crossed_circle', '03b6e0fcb3499374a867c041f52298f0',

        # Drag and drop
        'forbidden', 'no-drop', 'dnd-no-drop',
    ],
    'size_ns': ['size_ver', 'size-ver', 'ns-resize', 'n-resize', 's-resize', 'v_double_arrow', 'sb_v_double_arrow',
                'row-resize', 'split_v', 'double_arrow', '00008160000006810000408080010102',
                '2870a09082c103050810ffdffffe0204'],
    'size_ew': ['size_hor', 'size-hor', 'ew-resize', 'e-resize', 'w-resize', 'h_double_arrow', 'sb_h_double_arrow',
                'col-resize', 'split_h', '14fef782d02440884392942c11205230', '028006030e0e7ebffc7f7070c0600140'],
    'size_nwse': ['size_fdiag', 'size-fdiag', 'nwse-resize', 'nw-resize', 'se-resize', 'bd_double_arrow',
                  'c7088f0f3e6c8088236ef8e1e3e70000'],
    'size_nesw': ['size_bdiag', 'size-bdiag', 'nesw-resize', 'ne-resize', 'sw-resize', 'fd_double_arrow',
                  'fcf1c3c7cd4491d801f1e1c78f100000'],
    'move': [
        # Regular moves
        'fleur', 'size_all', 'all-scroll',

        # Drag and drop
        'move', 'grabbing', 'closedhand', 'dnd-move', 'dnd-none', 'dnd-ask', '4498f0e0c1937ffe01fd06f973665830',
        '9081237383d90e509aa00f00170e968f', 'fcf21c00b30f7e3f83fe0dfd12e71cff',
    ],
    'up_arrow': ['up-arrow', 'sb_up_arrow'],
    'link': ['pointer', 'pointing_hand', 'hand', 'hand1', 'hand2', '9d800788f1b08800ae810202380a0822',
             'e29285e634086352946a0e7090d73106'],
}



from pathlib import Path

from win2xcur.theme import CursorTheme, WIN_CURSORS
from win2xcur.writer import to_smart

INF_TEMPLATE = """\
; Right click on this file in Windows Explorer and select "Install".
; Then, run `main.cpl` and select the cursor theme "{name}".

[Version]
Signature = "$CHICAGO$"

[DefaultInstall]
CopyFiles = Scheme.Cur
AddReg    = Scheme.Reg

[DestinationDirs]
Scheme.Cur = 10,"%CUR_DIR%"

[Scheme.Reg]
HKLM,"SOFTWARE\\Microsoft\\Windows\\CurrentVersion\\Control Panel\\Cursors\\Schemes","%SCHEME_NAME%",,"{list}"

[Scheme.Cur]
{files}

[Strings]
CUR_DIR = "Cursors\\{name}"
SCHEME_NAME = "{name}"
{mapping}
"""


def export_windows_theme(theme: CursorTheme, directory: Path) -> None:
    files = []
    mapping = {}

    for name in WIN_CURSORS:
        cursor = getattr(theme, name)
        if cursor is None:
            continue

        ext, result = to_smart(cursor.frames)
        filename = f'{name}{ext}'

        with open(directory / filename, 'wb') as f:
            f.write(result)

        files.append(filename)
        mapping[name] = filename

    cursor_list = ','.join((f'%10%\\%CUR_DIR%\\%{name}%' if name in mapping else '') for name in WIN_CURSORS)
    inf_mapping = '\n'.join(f'{key} = "{value}"' for key, value in mapping.items())

    with open(directory / 'install.inf', 'w') as f:
        f.write(INF_TEMPLATE.format(name=theme.name, list=cursor_list, mapping=inf_mapping, files='\n'.join(files)))



import argparse

from win2xcur.cursor import CursorImage
from win2xcur.parser import open_blob


def print_images(images: list[CursorImage], prefix: str = '') -> None:
    for i, image in enumerate(images, 1):
        dimensions = f'{image.image.width}x{image.image.height}'
        print(f'{prefix}{i}. nominal size {image.nominal}, {dimensions}, hotspot: {image.hotspot}')


def main() -> None:
    parser = argparse.ArgumentParser(description='Converts Windows cursors to X11 cursors.')
    parser.add_argument('files', type=argparse.FileType('rb'), nargs='+',
                        help='Cursor files to inspect (*.cur, *.ani, or any X11 cursor)')
    args = parser.parse_args()

    for i, file in enumerate(args.files):
        if i:
            print()

        cursor = open_blob(file.read())
        print(f'Cursor file: {file.name}')

        if len(cursor.frames) > 1:
            for j, frame in enumerate(cursor.frames):
                print(f'  - Frame {j}, delay {frame.delay * 1000} ms')
                print_images(frame.images, prefix='    ')
        else:
            print_images(cursor.frames[0].images)


if __name__ == '__main__':
    main()



from pathlib import Path
from typing import Any

from win2xcur.parser import open_blob
from win2xcur.theme import CursorTheme, XCURSOR_ALIASES


def parse_xcursor_theme(name: str, directory: Path) -> CursorTheme:
    params: dict[str, Any] = {'name': name}

    for name, candidates in XCURSOR_ALIASES.items():
        for candidate in candidates:
            path = directory / candidate
            if path.is_file():
                with open(path, 'rb') as f:
                    params[name] = open_blob(f.read())
                break

    return CursorTheme(**params)



import argparse
import os
import shutil
import sys
from multiprocessing import cpu_count
from multiprocessing.pool import ThreadPool
from pathlib import Path

from win2xcur import scale, shadow
from win2xcur.parser.inf import parse_inf
from win2xcur.theme import XCURSOR_ALIASES
from win2xcur.writer import to_x11


def main() -> None:
    parser = argparse.ArgumentParser(description='Converts Windows cursor themes to X11 ones.')
    parser.add_argument('inf', type=Path, help='Windows cursor theme to convert (*.inf)')
    parser.add_argument('-o', '--output', '--output-dir', type=Path, default=os.curdir,
                        help='Directory to store converted cursor files.')
    parser.add_argument('-s', '--shadow', action='store_true',
                        help="Whether to emulate Windows's shadow effect")
    parser.add_argument('-O', '--shadow-opacity', type=int, default=50,
                        help='Opacity of the shadow (0 to 255)')
    parser.add_argument('-r', '--shadow-radius', type=float, default=0.1,
                        help='Radius of shadow blur effect (as fraction of width)')
    parser.add_argument('-S', '--shadow-sigma', type=float, default=0.1,
                        help='Sigma of shadow blur effect (as fraction of width)')
    parser.add_argument('-x', '--shadow-x', type=float, default=0.05,
                        help='x-offset of shadow (as fraction of width)')
    parser.add_argument('-y', '--shadow-y', type=float, default=0.05,
                        help='y-offset of shadow (as fraction of height)')
    parser.add_argument('-c', '--shadow-color', default='#000000',
                        help='color of the shadow')
    parser.add_argument('--scale', default=None, type=float,
                        help='Scale the cursor by the specified factor.')

    args = parser.parse_args()

    if not args.inf.is_file():
        sys.exit(f'INF file not found: {args.inf}')

    theme = parse_inf(args.inf)

    if not theme.arrow:
        sys.exit(f'Basic pointer cursor not found in theme: {args.inf}')

    def process(name: str, aliases: list[str]) -> None:
        cursor = getattr(theme, name)
        if cursor is None:
            return

        if args.scale:
            scale.apply_to_frames(cursor.frames, scale=args.scale)

        if args.shadow:
            shadow.apply_to_frames(cursor.frames, color=args.shadow_color, radius=args.shadow_radius,
                                   sigma=args.shadow_sigma, xoffset=args.shadow_x, yoffset=args.shadow_y)

        result = to_x11(cursor.frames)
        canonical = aliases[0]
        with open(args.output / canonical, 'wb') as f:
            f.write(result)

        for alias in aliases[1:]:
            if os.name == 'posix':
                output = args.output / alias
                try:
                    os.symlink(canonical, output)
                except FileExistsError:
                    os.remove(output)
                    os.symlink(canonical, output)
            else:
                shutil.copyfile(args.output / canonical, args.output / alias)

    with ThreadPool(cpu_count()) as pool:
        pool.starmap(process, XCURSOR_ALIASES.items())


if __name__ == '__main__':
    main()



import argparse
import os
import sys
from multiprocessing import cpu_count
from multiprocessing.pool import ThreadPool
from pathlib import Path

from win2xcur import scale
from win2xcur.parser.xtheme import parse_xcursor_theme
from win2xcur.theme import ALL_CURSORS
from win2xcur.writer.inf import export_windows_theme


def main() -> None:
    parser = argparse.ArgumentParser(description='Converts X11 cursor themes to Windows ones.')
    parser.add_argument('source', type=Path, help='X11 cursor files directory')
    parser.add_argument('-n', '--name', required=True, help='name of the cursor theme')
    parser.add_argument('-o', '--output', '--output-dir', type=Path, default=os.curdir,
                        help='Directory to store converted cursor files and install.inf.')
    parser.add_argument('-S', '--scale', default=None, type=float,
                        help='Scale the cursor by the specified factor.')

    args = parser.parse_args()

    if not args.source.is_dir():
        sys.exit(f'Source must be a directory: {args.source}')

    theme = parse_xcursor_theme(args.name, args.source)

    if not theme.arrow:
        sys.exit(f'Basic pointer cursor not found in theme: {args.source}')

    def process(name: str) -> None:
        cursor = getattr(theme, name)
        if args.scale:
            scale.apply_to_frames(cursor.frames, scale=args.scale)

    with ThreadPool(cpu_count()) as pool:
        pool.map(process, ALL_CURSORS)

    args.output.mkdir(exist_ok=True)
    export_windows_theme(theme, args.output)


if __name__ == '__main__':
    main()
